/* AUTOR: RAFAEL ABERTO MORENO PARRA

   Evaluador de expresiones en TURBO C.	Mayo 13 de 1993.

   Sus caracterisiticas principales son:
   - Evaluador NO recursivo.
   - Optimizado para ciclos.
   - Evalua funciones trigonomtricas en radianes.

   Forma de uso:

	 double capturaecuacion(char *expresin, double valor_x, double valor_y);

	 Donde *expresin es una cadena donde reposa la expresin a evaluar,
	 por ejemplo: "4*2-x+y*3".

	 valor_x y valor_y son los valores con que va las variables 'x' y 'y'
	 en la expresin.

	 Esta subrutina retorna el valor de la expresin.

	 Ej:
		 void main()
		 {
		  double resultado;
		  unsigned char funcion[]="3*4+y-21*x";
		  double x=10.4, y=23.5;

		  resultado=capturaecuacion(funcion, x, y);

		  printf("el resultado es %f", resultado);
		 }


	 double comoevalua(double valor_x, double valor_y);



	 retorna los valores de una expresin anteriormente evaluada,
	 es muy veloz para ciclos.

	 Ej:

		 void main()
		 {
		  double resultado;
		  unsigned char funcion[]="3*4+y-21*x";
		  double x=10.4, y=23.5;
		  unsigned int contador;

		  resultado=capturaecuacion(funcion, x, y);

		  printf("el resultado es %f\n", resultado);

		  for(i=1; i<=60; i++)
		  {
		   x*=i; y+=i;
		   resultado=comoevalua(x, y);
		   printf("resultados parciales %f\n", resultado);
		  }
		 }



	 unsigned int corrige_sintaxis(unsigned char *expresion)

	 retorna 0 si NO hay errores en la expresin.
	 retorna
		 1:  Dobles operandos. Ej: 5*+3
		 2:  Parentesis desbalanceados  Ej: 4-(3*4
		 3:  Numero seguido de (      Ej: 7-4(3*2)
		 4:  Operando(+,-,*,/) seguido de )   Ej: 3-(4*)+2
		 5:  Letra seguida de )     Ej: 7-(5-a)
		 6:  Mandato Desconocido  Ej: flow(3*4)

*/


/* COMO FUE PROGRAMADO:

   Supongamos una cadena de expresin as:
	13-4*sin(3+x)+8-(12+x*y)

   la expresin se desintegra en acums:

   acum[0]="12+x*y"             => 13-4*sin(3+x)+8-acum0;
   acum[1]="3x+8"               => 13-4*sinacum1+8-acum0;
   acum[2]="sinacum1"           => 13-4*acum2+8-acum0;
   acum[3]="13-4*acum2+8-acum0"

   cada 'acum' es evaluado como una expresin simple.
   acum[0]="12+x*y"
   => vector_numeros[0][0]="12"  vector_operador[0][0]='+';
	  vector_numeros[0][1]="x"   vector_operador[0][1]='*';
	  vector_numeros[0][2]="y"   vector_operador[0][2]='F';
					 
	  acum 

*/

#include <stdio.h>
#include <string.h>
#include <alloc.h>
#include <ctype.h>
#include <math.h>
#include <stdlib.h>

#include "cadena.h"

unsigned char acum[30][50];
unsigned char num_acums; /* variable que cuenta el numero de acums */
unsigned char vector_numeros[27][20][10];
unsigned char vector_operador[27][27];
double acum_float[30];
unsigned char MA_ERROR=0;


unsigned char funciones[19][16]={ "asna", "arcosenoa", "acsa",
				"arcocosenoa", "atna", "arcotangentea",
				"cosa", "cosenoa", "sina", "sena", "senoa",
				"tana", "tangentea", "expa", "exponenciala",
				"absa", "sinha", "cosha", "tanha" };

double capturaecuacion(unsigned char *, double, double);
void optimizador(unsigned char *);
void reemplazar_expr(unsigned char *, unsigned char *, unsigned char, unsigned char);
void ver_funciones (unsigned char *);
void arregla_acumn(void);
void ciclocumn(void);
void desintegra_acumn(unsigned char *, unsigned char);
double comoevalua(double, double);
double evalua_expresion(unsigned char, double, double);



/* FUNCIONES DE SINTAXIS */
unsigned int corrige_sintaxis(unsigned char *);


double capturaecuacion(register unsigned char *p, double x, double y)
{
 register unsigned char *p2, k;

 num_acums=0;
 for(k=0; k<=29; k++) acum[k][0]='\0';
 cad_minusculas(p);
 eliminaespacios(p);

 p2=(unsigned char *)malloc(strlen(p)+4);
 strcpy(p2, "@\0");
 strcat(p2, p); /* la expresin debe empezar siempre en @ */

 optimizador(p2); /* desglosa la cadena(expresion algebraica) en acums */
 arregla_acumn();
 ciclocumn();
 free(p2);
 return(comoevalua(x, y));
}

void optimizador(unsigned char *p)
{
 /* convierte la expresion algebraica en acums */
 register unsigned char i,j, *subexpresion, *reemplazo;

 subexpresion=(unsigned char *) malloc(strlen(p)+5);
 reemplazo=(unsigned char *) malloc(4);

 do{
  for(i=strlen(p); i>=1; i--)
   if( *(p+i) == '(' ) /* busca parentesis */
   {
	*(p+i)='<'; /* los cambia por < */
    break;
   }
  if(i) /* si hay parentesis ( se debe buscar el ) */
  {
   for(j=i; *(p+j); j++)
    if( *(p+j) == ')')
    {
     *(p+j)='>'; /* cambia el ) por > */
     mid(subexpresion,p,i+2,j); /* extrae la expresion entre parentesis */
     strcpy(acum[num_acums], subexpresion);

     /* reemplaza la subexpresion entre parentesis por un a1, a2, a3, .. */
	 reemplazo[0]='a';
	 reemplazo[1]=num_acums+65;
	 reemplazo[2]='\0';
     reemplazar_expr(p, reemplazo, i, j+2);
     num_acums++;

	ver_funciones(p);
	break;
    }
  }
 }while(i); /* para hasta que se acaben los parentesis */
 ver_funciones(p); /* ve si hay una funcion en expresion sin parentesis */
 strcpy(acum[num_acums], p);
 free(reemplazo); free(subexpresion);
}

void reemplazar_expr(unsigned char *p, unsigned char *reemplazo,
		     unsigned char desde, unsigned char hasta)
{
 register unsigned char *tempor, *tempor2;

 tempor=(unsigned char *)malloc(strlen(p)+2);
 tempor2=(unsigned char *)malloc(strlen(p)+2);

 left(tempor, p, desde);
 strcat(tempor, reemplazo);
 right(tempor2, p, hasta);
 strcat(tempor, tempor2);
 strcpy(p, tempor);

 free(tempor);
 free(tempor2);
}

void ver_funciones (unsigned char *p)
{
 /* esta subrutina chequea funciones tipo seno, coseno, etc.... */
 unsigned char *identificacion="AABBCCDDEEEFFGGHIJK";
 unsigned char *p_funcion, funcion=1;
 register unsigned char i, *t;
 unsigned char *tempor, *tempor2, *reemplazo;
 unsigned int contador=0;



 tempor =(unsigned char *) malloc(strlen(p)+2);
 tempor2=(unsigned char *) malloc(strlen(p)+2);
 reemplazo=(unsigned char *) malloc(strlen(p)+2);


 while( (p_funcion=strstr(p, funciones[contador])) == NULL)
 {
  if(contador>=18){ funcion=0; break; }
  contador++;
 }
 if(funcion) /* encontro una funcion */
 {
  t=p;
  for(i=0; *t++; i++)
   if( (strncmp(t, funciones[contador], strlen(funciones[contador]))) == 0 )
   {
	left(tempor, p, i+1);
	right(tempor2, p, i+strlen(funciones[contador])+3);
	acum[num_acums][0]=identificacion[contador];

	reemplazo[0]='a';
	reemplazo[1]=num_acums+65;
	reemplazo[2]='\0';
	strcat(tempor, reemplazo);
	strcat(tempor, tempor2);
	strcpy(p, tempor);
	num_acums++;
   }
  }
 free(tempor); free(tempor2); free(reemplazo);
}

/* subrutina que arregla los acums */
void arregla_acumn()
{
 register unsigned int k;
 register unsigned char *tempor;

 tempor=(unsigned char *)malloc(strlen(acum[num_acums])+3);
 right(tempor, acum[num_acums], 2);
 strcpy(acum[num_acums], tempor);
 free(tempor);

 for(k=0; k<=num_acums; k++)
  if(acum[k][0]=='-')  /* cualquier expresion negativa ej: -x*3 => 0-x*3 */
  {
   tempor=(unsigned char *)malloc(strlen(acum[k])+3);
   strcpy(tempor, "0\0");
   strcat(tempor, acum[k]);
   strcpy(acum[k], tempor);
   free(tempor);
  }
}

void ciclocumn()
{
 register unsigned char i;

 for(i=0; i<=num_acums; i++)
  if(acum[i][0]<'A' || acum[i][0]>'Z') desintegra_acumn(acum[i], i);
}



void desintegra_acumn(unsigned char *expresion, unsigned char indicadorp)
/* expresion  expresin algebraica simple
   indicadorp; cual acum */
{
 register unsigned char ciclo, caracter, ij=0;
 unsigned char *elnumero, j=0, tamano;

 elnumero=(unsigned char *)malloc(strlen(expresion)+2);
 *elnumero='\0';
 for(ciclo=0; ciclo<=strlen(expresion); ciclo++)
  switch( caracter=*(expresion+ciclo) )
  {
   case 'x': /* variable */
   case 'y':
			vector_numeros[indicadorp][ij][0]=caracter;
			ij++;
			break;

   case 'a': /* es un acumn */
			 vector_numeros[indicadorp][ij][0]= *(expresion+ciclo+1);
			 ij++;
			 ciclo+=1;
			 break;
   case '*':
   case '/':
   case '+':
   case '-':
   case '^':
			if(elnumero[0]!='\0')
			{
			 strcpy(vector_numeros[indicadorp][ij], elnumero);
			 ij++;
			 elnumero[0]='\0';
			}
			vector_operador[indicadorp][j++]=caracter;
			break;
   default: /* o sea un numero */
			tamano=strlen(elnumero);
			elnumero[tamano]=caracter;
			elnumero[tamano+1]='\0';
   } /* for */
   if( *elnumero != '\0') strcpy(vector_numeros[indicadorp][ij], elnumero);
   vector_operador[indicadorp][j]='F';
   free(elnumero);
}


double comoevalua(double x, double y)
{
 register unsigned char indicador;

 for(indicador=0; indicador<=num_acums; indicador++)
  switch(acum[indicador][0]){
   case 'A': acum_float[indicador]=asin(acum_float[indicador-1]);
			 break;
   case 'B': acum_float[indicador]=acos(acum_float[indicador-1]);
			 break;
   case 'C': acum_float[indicador]=atan(acum_float[indicador-1]);
			 break;
   case 'D': acum_float[indicador]=cos(acum_float[indicador-1]);
			 break;
   case 'E': acum_float[indicador]=sin(acum_float[indicador-1]);
			 break;
   case 'F': acum_float[indicador]=tan(acum_float[indicador-1]);
			 break;
   case 'G': acum_float[indicador]=exp(acum_float[indicador-1]);
			 break;
   case 'H': acum_float[indicador]=fabs(acum_float[indicador-1]);
			 break;
   case 'I': acum_float[indicador]=sinh(acum_float[indicador-1]);
			 break;
   case 'J': acum_float[indicador]=cosh(acum_float[indicador-1]);
			 break;
   case 'K': acum_float[indicador]=tanh(acum_float[indicador-1]);
			 break;
   default:  acum_float[indicador]=evalua_expresion(indicador, x, y);
  }
  return(acum_float[num_acums]);
}



double evalua_expresion(unsigned char indicadorp, double x, double y)
{
 unsigned char operando[20];
 double numero[20], acumula=0;
 register unsigned char i, j, ij, lim_array=0;

 for(ij=0;;ij++)
 {
  switch(vector_numeros[indicadorp][ij][0])
  {
   case 'x': numero[ij]=x; break;
   case 'y': numero[ij]=y; break;
   case '.':
   case '1':
   case '2':
   case '3':
   case '4':
   case '5':
   case '6':
   case '7':
   case '8':
   case '9':
   case '0':
			 numero[ij]=atof(vector_numeros[indicadorp][ij]);
			 break;
   default:	 numero[ij]=acum_float[vector_numeros[indicadorp][ij][0]-65];
  }
  operando[ij]=vector_operador[indicadorp][ij];
  if(operando[ij]=='F') break;
 }
 lim_array=ij;

 for(i=0; i<=lim_array; i++)
  while(operando[i]=='^')
  {
   numero[i]=pow(numero[i], numero[i+1]);
   for(j=i+2; j<=lim_array+1; j++)
   {
	numero[j-1]=numero[j];
	operando[j-2]=operando[j-1];
   }
   lim_array--;
  }

 for(i=0; i<=lim_array; i++)
  while(operando[i]=='/')
  {
   if(numero[i+1]!=0) numero[i]=numero[i]/numero[i+1];
   else { MA_ERROR=1; numero[i]=0; }
   for(j=i+2; j<=lim_array+1; j++)
   {
	numero[j-1]=numero[j];
	operando[j-2]=operando[j-1];
   }
   lim_array--;
  }

 for(i=0; i<=lim_array; i++)
  while(operando[i]=='*')
  {
   numero[i]=numero[i]*numero[i+1];
   for(j=i+2; j<=lim_array+1; j++)
   {
	numero[j-1]=numero[j];
	operando[j-2]=operando[j-1];
   }
   lim_array--;
  }

 acumula=numero[0];
 for(i=0; i<=lim_array; i++)
  if(operando[i]=='+') acumula+=numero[i+1];
  else
   if(operando[i]=='-') acumula-=numero[i+1];

 return acumula;
}


/* ********************************** */
/* FUNCIONES DE CORRECION DE SINTAXIS */
/* ********************************** */

/* RETORNA:
	1:  Dobles operandos. Ej: 5*+3
	2:  Parentesis desbalanceados  Ej: 4-(3*4
	3:  Numero seguido de (      Ej: 7-4(3*2)
	4:  Operando(+,-,*,/) seguido de )   Ej: 3-(4*)+2
	5:  Letra seguida de )     Ej: 7-(5-a)
	6:  Mandato Desconocido  Ej: flow(3*4)
*/

unsigned int corrige_sintaxis(unsigned char *p)
{
 unsigned int k;
 unsigned int parentesis=0;
 unsigned char cad_temp[150], cad_temp2[150];

 unsigned char *p_funcion, contador=0, funcion=1;
 register unsigned char i, *t;
 unsigned char *tempor, *tempor2, reemplazo[4];

 strcpy(cad_temp, p);
 cad_minusculas(cad_temp);
 eliminaespacios(cad_temp);

 for(k=0; *(cad_temp+k); k++)
 {
  if( cad_temp[k]=='^' || cad_temp[k]=='-' || cad_temp[k]== '*' || cad_temp[k]=='+')
   if( cad_temp[k+1]=='^' || cad_temp[k+1]=='-' || cad_temp[k+1]== '*' || cad_temp[k+1]=='+')
	return 1;

  if( cad_temp[k]>='0' && cad_temp[k]<='9' && cad_temp[k+1]=='(')
   return 3;

  if( cad_temp[k]=='+' || cad_temp[k]=='-' || cad_temp[k]=='*' || cad_temp[k]=='/')
   if( cad_temp[k+1]==')')
	return 4;

  /* la siguiente comprobacion solo es para agenda,
	 en el real tanto 'x' como 'y' pueden estar (4-x) */
  if( cad_temp[k]>='a' && cad_temp[k]<='z' && cad_temp[k+1]==')')
   return 5;

  if( cad_temp[k]=='(' ) parentesis++;
  if( cad_temp[k]==')' ) parentesis--;
 }
 if(parentesis) return 2;

 /* **************************************************** */
 /* chequea que esten solo las funciones trigonometricas */
 /* **************************************************** */
 contador=0;
 strcpy(cad_temp, p);
 cad_minusculas(cad_temp);
 eliminaespacios(cad_temp);
 strcpy(cad_temp2, " \0");
 strcat(cad_temp2, cad_temp);
 strcpy(cad_temp, cad_temp2);
 p_funcion=NULL;


 tempor =(unsigned char *) malloc(strlen(cad_temp)+2);
 tempor2=(unsigned char *) malloc(strlen(cad_temp)+2);

 while(funcion)
 {
  while(p_funcion==NULL)
  {
   strcpy(cad_temp2, funciones[contador]);
   cad_temp2[strlen(cad_temp2)-1]='(';

   p_funcion=strstr(cad_temp, cad_temp2);
   if(p_funcion!=NULL) break;

   if(contador>=18){ funcion=0; break; }
   contador++;
  }
  if(funcion) /* encontro una funcion */
  {
   contador=0;
   p_funcion=NULL;
   t=cad_temp;
   for(i=0; *t++; i++)
	if( (strncmp(t, cad_temp2, strlen(cad_temp2))) == 0 )
	{
	 left(tempor, cad_temp, i+1);
	 right(tempor2, cad_temp, i+strlen(funciones[contador])+2);

	 reemplazo[0]='$';
	 reemplazo[1]='\0';
	 strcat(tempor, reemplazo);
	 strcat(tempor, tempor2);
	 strcpy(cad_temp, tempor);
	}
   }
 } /* while(funcion) */

 free(tempor); free(tempor2);
 /* si queda alguna letra, hay un error */
 for(k=0; *(cad_temp+k); k++)
  if( cad_temp[k]>='a' && cad_temp[k]<='z') return 6;

 return 0; /* cadena OK! */
}


int matherr(struct exception *a)
{
   MA_ERROR=1;
   return 1;
}
